@@ -1,12 +1,14 @@
+import logging
 import numpy as np
 import pandas as pd
+import joblib
 import astropy.units as au
-from astropy.units import AA
 from lmfit.models import PolynomialModel
 from sys import exit
 from pathlib import Path
+from scipy import signal
 
-from lime import _logger
+_logger = logging.getLogger('LiMe')
 
 try:
     from specutils import Spectrum1D, SpectralRegion
@@ -23,17 +25,33 @@ SYB_LIST = ["M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I
 ASTRO_UNITS_KEYS = {'A': au.AA,
                     'um': au.um,
                     'nm': au.nm,
-                    'flam': au.erg / au.s / au.cm ** 2 / au.AA}
-
-WAVE_UNITS_DEFAULT, FLUX_UNITS_DEFAULT = au.AA, au.erg / au.s / au.cm ** 2 / au.AA
+                    'Hz': au.Hz,
+                    'cm': au.cm,
+                    'mm': au.mm,
+                    'Flam': au.erg/au.s/au.cm**2/au.AA,
+                    'Fnu': au.erg/au.s/au.cm**2/au.Hz,
+                    'Jy': au.Jy,
+                    'mJy': au.mJy}
 
 UNITS_LATEX_DICT = {'A': '\AA',
                     'um': '\mu\!m',
                     'nm': 'nm',
-                    'erg/cm^2/s/A': r'erg\,cm^{-2} s^{-1} \AA^{-1}',
+                    'Hz': 'Hz',
+                    'cm': 'cm',
+                    'mm': 'mm',
+                    'Flam': r'erg\,cm^{-2}s^{-1}\AA^{-1}',
+                    'Fnu': r'erg\,cm^{-2}s^{-1}\Hz^{-1}',
                     'Jy': 'Jy',
                     'mJy': 'mJy'}
 
+DISPERSION_UNITS = ('A', 'um', 'nm', 'Hz', 'cm', 'mm')
+
+FLUX_DENSITY_UNITS = ('Flam', 'Fnu', 'Jy', 'mJy')
+
+WAVE_UNITS_DEFAULT, FLUX_UNITS_DEFAULT = au.AA, au.erg / au.s / au.cm ** 2 / au.AA
+
+MACHINE_PATH = Path(__file__).parent/'resources/LogitistRegression_v2_cost1_logNorm.joblib'
+
 #[Y erg/cm^2/s/A] = 2.99792458E+21 * [X1 W/m^2/Hz] / [X2 A]^2
 # 2.99792458E+17 units
 # 1 Jy = 10−26 W⋅m−2⋅Hz−1
@@ -270,15 +288,23 @@ def latex_science_float(f, dec=2):
         return float_str
 
 
-def unit_convertor(in_units, out_units, sig_fig=None):
+def unit_convertor(in_units, out_units, wave_array=None, flux_array=None, dispersion_units=None, sig_fig=None):
 
-    unit_factor = 1 * ASTRO_UNITS_KEYS[in_units]
-    unit_factor = unit_factor.to(ASTRO_UNITS_KEYS[out_units])
+    if (in_units in DISPERSION_UNITS) and (out_units in DISPERSION_UNITS):
+        input_array = wave_array * ASTRO_UNITS_KEYS[in_units]
+        output_array = input_array.to(ASTRO_UNITS_KEYS[out_units])
+
+    elif (in_units in FLUX_DENSITY_UNITS) and (out_units in FLUX_DENSITY_UNITS):
+        input_array = flux_array * ASTRO_UNITS_KEYS[in_units]
+        wave_array = wave_array * ASTRO_UNITS_KEYS[dispersion_units]
+        output_array = input_array.to(ASTRO_UNITS_KEYS[out_units], au.spectral_density(wave_array))
+    else:
+        _logger.warning(f'Input units {in_units} could not be converted to {out_units}')
 
     if sig_fig is None:
-        return unit_factor.value
+        return output_array.value
     else:
-        return np.round(unit_factor.value, sig_fig)
+        return np.round(output_array.value, sig_fig)
 
 
 def spectral_mask_generator(wave_interval=None, lines_list=None, ion_list=None, z_range=None,
@@ -303,7 +329,7 @@ def spectral_mask_generator(wave_interval=None, lines_list=None, ion_list=None,
     # Convert the output mask to the units requested by the user
     if units_wave != 'A':
 
-        conversion_factor = unit_convertor('A', units_wave, sig_fig)
+        conversion_factor = unit_convertor('A', units_wave,  wave_array=1, dispersion_units='dispersion axis', sig_fig=sig_fig)
         mask_df.loc[:, 'w1':'w6'] = mask_df.loc[:, 'w1':'w6'] * conversion_factor
 
         wave_array = wave_array * conversion_factor if sig_fig is None else np.round(wave_array * conversion_factor, sig_fig)
@@ -340,14 +366,133 @@ def spectral_mask_generator(wave_interval=None, lines_list=None, ion_list=None,
     return output_mask
 
 
-
-
 class LineFinder:
 
-    def __init__(self):
+    def __init__(self, machine_model_path=MACHINE_PATH):
+
+        self.ml_model = joblib.load(machine_model_path)
 
         return
 
+    def continuum_fitting(self, degree_list=[3, 7, 7, 7], threshold_list=[5, 3, 2, 2], plot_results=False,
+                          return_std=False):
+
+        # Check for a masked array
+        if np.ma.is_masked(self.flux):
+            mask_cont = ~self.flux.mask
+            input_wave, input_flux = self.wave.data, self.flux.data
+        else:
+            mask_cont = np.ones(self.flux.size).astype(bool)
+            input_wave, input_flux = self.wave, self.flux
+
+        # Loop throught the fitting degree
+        for i, degree in enumerate(degree_list):
+
+            # Establishing the flux limits
+            low_lim, high_lim = np.percentile(input_flux[mask_cont], (16, 84))
+            low_lim, high_lim = low_lim / threshold_list[i], high_lim * threshold_list[i]
+
+            # Add new entries to the mask
+            mask_cont = mask_cont & (input_flux >= low_lim) & (input_flux <= high_lim)
+
+            poly3Mod = PolynomialModel(prefix=f'poly_{degree}', degree=degree)
+            poly3Params = poly3Mod.guess(input_flux[mask_cont], x=input_wave[mask_cont])
+
+            try:
+                poly3Out = poly3Mod.fit(input_flux[mask_cont], poly3Params, x=input_wave[mask_cont])
+                continuum_fit = poly3Out.eval(x=input_wave)
+
+            except TypeError:
+                _logger.warning(f'- The continuum fitting polynomial as degree ({degree}) is larger than data points'
+                                f' number')
+                continuum_fit = np.full(input_wave.size, np.nan)
+
+            # Compute the continuum and assign replace the value outside the bands the new continuum
+            if plot_results:
+                title = f'Continuum fitting, iteration ({i+1}/{len(degree_list)})'
+                continuum_full = poly3Out.eval(x=self.wave.data)
+                self._plot_continuum_fit(continuum_full, mask_cont, low_lim, high_lim, threshold_list[i], title)
+
+        # Include the standard deviation of the spectrum for the unmasked pixels
+        if return_std:
+            std_spec = np.std((self.flux-continuum_fit)[mask_cont])
+            output_params = (continuum_fit, std_spec)
+        else:
+            output_params = continuum_fit
+
+        return output_params
+
+    def peak_detection(self, limit_threshold=None, continuum=None, distance=4, ml_mask=None, plot_results=False):
+
+        # No user imput provided compute the intensity threshold from the 84th percentil
+        limit_threshold = np.percentile(self.flux, 84) if limit_threshold is None else limit_threshold
+        limit_threshold = limit_threshold + continuum if continuum is not None else limit_threshold
+
+        # Index the intensity peaks
+        # input_flux = self.flux.data if np.ma.is_masked()
+        # mask_valid = ~self.flux.mask if np.ma.is_masked(self.flux) else np.ones(self.flux.data.size).astype(bool)
+        # peak_fp, _ = signal.find_peaks(self.flux.data[mask_valid], height=limit_threshold[mask_valid], distance=distance)
+
+        peak_fp, _ = signal.find_peaks(self.flux, height=limit_threshold, distance=distance)
+
+        # Plot the results
+        if plot_results:
+            self._plot_peak_detection(peak_fp, limit_threshold, continuum, ml_mask=ml_mask,
+                                      plot_title='Peak detection results ')
+
+        return peak_fp
+
+    def ml_line_detection(self, continuum, box_width=11):
+
+        # Normalize the flux
+        input_flux = self.flux if not np.ma.is_masked(self.flux) else self.flux.data
+        input_flux = np.log10((input_flux/continuum - 1) + 10)
+
+        # Reshape to the training dimensions
+        input_flux = np.array(input_flux, ndmin=2)
+
+        # Container for the true pixels
+        detection_mask = np.zeros(self.flux.size).astype(bool)
+
+        # Case of 1D
+        spectrum_pixels = input_flux.shape[1]
+        for i in np.arange(spectrum_pixels):
+            if i + box_width <= spectrum_pixels:
+                y = input_flux[:, i:i + box_width]
+                if not np.any(np.isnan(y)):
+                    detection_mask[i:i + box_width] = detection_mask[i:i + box_width] | self.ml_model.predict(y)[0]
+                    # print(f'y {i} ({np.sum(y)}): {self.ml_model.predict(y)[0]}')
+
+        return detection_mask
+
+    def line_detection(self, poly_degree=[3, 7, 7, 7], emis_threshold=[5, 3, 2, 2], noise_sigma_factor=3, lines_log=None,
+                       line_type='emission', width_tol=5, width_mode='fixed', ml_detection=False, plot_cont_calc=False,
+                       plot_peak_calc=True):
+
+        # Fit the continuum
+        cont_flux, cond_Std = self.continuum_fitting(degree_list=poly_degree, threshold_list=emis_threshold,
+                                                     return_std=True, plot_results=plot_cont_calc)
+
+        # Check via machine learning algorithm
+        ml_mask = self.ml_line_detection(cont_flux) if ml_detection else None
+
+        # Check for the peaks of the emission lines
+        detec_min = noise_sigma_factor * cond_Std
+        idcs_peaks = self.peak_detection(detec_min, cont_flux, plot_results=plot_peak_calc, ml_mask=ml_mask)
+
+        # Compare against the theoretical values
+        if lines_log is not None:
+
+            # Match peaks with theoretical lines
+            matched_DF = self.label_peaks(idcs_peaks, lines_log, width_tol=width_tol, width_mode=width_mode,
+                                          line_type=line_type)
+
+            return matched_DF
+
+        else:
+
+            return
+
     def match_line_mask(self, log, noise_region, detect_threshold=3, emis_threshold=(4, 4), abs_threshold=(1.5, 1.5),
                         poly_degree=(3, 7), width_tol=5, line_type='emission', width_mode='fixed'):
 
@@ -488,17 +633,26 @@ class LineFinder:
     def label_peaks(self, peak_table, mask_df, line_type='emission', width_tol=5, width_mode='auto', detect_check=False):
 
         # TODO auto param should be changed to boolean
+        # Establish the type of input values for the peak indexes, first numpy array
+        if isinstance(peak_table, np.ndarray):
+            idcsLinePeak = peak_table
 
+        # Specutils table
+        else:
+            # Query the lines from the astropy finder tables #
+            if len(peak_table) != 0:
+                idcsLineType = peak_table['line_type'] == line_type
+                idcsLinePeak = np.array(peak_table[idcsLineType]['line_center_index'])
+            else:
+                idcsLinePeak = np.array([])
+
+        # Security check in case no lines detected
+        if len(idcsLinePeak) == 0:
+            return pd.DataFrame(columns=mask_df.columns)
+
+        # Prepare dataframe to stored the matched lines
         matched_DF = pd.DataFrame.copy(mask_df)
-
-        # Security check for case no lines detected
-        if len(peak_table) == 0:
-            return pd.DataFrame(columns=matched_DF.columns)
-
-        # Query the lines from the astropy finder tables #
-        idcsLineType = peak_table['line_type'] == line_type
-        idcsLinePeak = np.array(peak_table[idcsLineType]['line_center_index'])
-        wave_peaks = self.wave_rest[idcsLinePeak]
+        matched_DF['signal_peak'] = np.nan
 
         # Theoretical wave values
         ion_array, waveTheory, latexLabel_array = label_decomposition(matched_DF.index.values, units_wave=self.units_wave)
@@ -506,9 +660,17 @@ class LineFinder:
 
         # Match the lines with the theoretical emission
         tolerance = np.diff(self.wave_rest).mean() * width_tol
-        matched_DF['observation'] = 'not detected'
+        matched_DF['observation'] = 'not_detected'
         unidentifiedLine = dict.fromkeys(matched_DF.columns.values, np.nan)
 
+        # Get the wavelength peaks
+        wave_peaks = self.wave_rest[idcsLinePeak]
+
+        # Only treat pixels outisde the masks
+        if np.ma.is_masked(wave_peaks):
+            idcsLinePeak = idcsLinePeak[~wave_peaks.mask]
+            wave_peaks = wave_peaks[~wave_peaks.mask].data
+
         for i in np.arange(wave_peaks.size):
 
             idx_array = np.where(np.isclose(a=waveTheory, b=wave_peaks[i], atol=tolerance))
@@ -527,6 +689,7 @@ class LineFinder:
 
                 row_index = matched_DF.index[matched_DF.wavelength == waveTheory[idx_array[0][0]]]
                 matched_DF.loc[row_index, 'observation'] = 'detected'
+                matched_DF.loc[row_index, 'signal_peak'] = idcsLinePeak[i]
                 theoLineLabel = row_index[0]
 
                 blended_check = True if '_b' in theoLineLabel else False
@@ -542,15 +705,11 @@ class LineFinder:
                         matched_DF.loc[row_index, 'w4'] = self.wave_rest[idx_max]
 
         # Include_only_detected
-        idcs_unknown = matched_DF['observation'] == 'not detected'
+        idcs_unknown = matched_DF['observation'] == 'not_detected'
         matched_DF.drop(index=matched_DF.loc[idcs_unknown].index.values, inplace=True)
 
         # Sort by wavelength
         matched_DF.sort_values('wavelength', inplace=True)
         matched_DF.drop(columns=['wavelength', 'observation'], inplace=True)
 
-        # Latex labels
-        # ion_array, wavelength_array, latexLabel_array = label_decomposition(matched_DF.index.values)
-        # matched_DF['latexLabel'] = latexLabel_array
-
-        return matched_DF
\ No newline at end of file
+        return matched_DF
